import atexit
import json
import logging
from statistics import mean
from typing import Callable, Dict
import os

from ruxit.api.base_plugin import BasePlugin
from ruxit.api.selectors import HostSelector
from ruxit.api.data import PluginProperty, MEAttribute

from perfmon_api import Perfmon
from process_controller import restart_citrix_agent, check_if_citrix_agent_running, kill_citrix_agent
from log_proxy import get_lines_from_file

current_file_path = os.path.dirname(os.path.realpath(__file__))
logger = logging.getLogger(__name__)


def count(my_list):
    # Remove one element because of _ServerTotal instance
    length = len(my_list)
    if length == 0:
        return 0
    return length - 1


aggregation_methods = {"sum": sum, "avg": mean, "count": count, "instance": None}


class CitrixPlugin(BasePlugin):
    def initialize(self, **kwargs):
        self.metrics = kwargs["json_config"]["metrics"]
        self.metric_methods: Dict[str, Callable] = {
            "absolute": self.results_builder.absolute,
            "relative": self.results_builder.relative,
            "per_second": self.results_builder.per_second,
        }

        self.log_path = os.path.join(current_file_path, "citrixagent", "log", "citrix-agent.log")
        self.update_config()
        if self.config.get("citrixagent_run") == "Enabled":
            restart_citrix_agent()

        self.times_not_logged_lines = 0

        atexit.register(kill_citrix_agent)

    def query(self, **kwargs):

        rum_enabled = self.config.get("citrixagent_run") == "Enabled"

        self._report_property("Citrix agent version", kwargs["json_config"]["version"])

        # Log lines from citrix-agent.log
        self.process_log()

        # Check if citrix-agent.exe is running, and start it if necessary
        if rum_enabled:
            check_if_citrix_agent_running()

            if self.times_not_logged_lines >= 5:
                self.logger.warning("The citrix agent has not logged anything for 5 minutes, restarting")
                restart_citrix_agent()
                self.times_not_logged_lines = 0

        else:
            kill_citrix_agent()

        # Report vda properties
        self.report_vda_properties()

        # Send all perfmon counters
        self.report_perfmon_counters()

    def report_perfmon_counters(self):
        metric_collection = self.config.get("monitoring_mode", "Basic")

        # Filter only metrics that match the monitoring_mode
        metrics_to_collect = [metric for metric in self.metrics if metric_collection in metric["source"]["modes"]]

        logger.info(f"Metric Collection mode: {metric_collection}. Will attempt to collect {len(metrics_to_collect)} metrics")

        with Perfmon(logger) as p:
            for metric in metrics_to_collect:

                metric_type = metric["source"].get("type", "absolute")
                metric_dimensions = metric["timeseries"].get("dimensions", [])
                p.add_counter(
                    metric["source"]["object"],
                    metric["source"]["counter"],
                    metric["timeseries"]["key"],
                    metric["source"]["perfmon_aggregation"],
                    metric_type,
                    dimension_name=metric_dimensions[0] if metric_dimensions else None,
                )
            p.collect_data()

            for key, counters in p.get_counters_per_key().items():
                if counters:
                    aggregation_method = aggregation_methods[counters[0].aggregation]
                    metric_type = counters[0].metric_type
                    if aggregation_method is None:
                        for counter in counters:
                            logger.info(f"Reporting {counter.instance}={counter.value} for key {key}.")
                            self.metric_methods[metric_type](
                                key=key, value=counter.value, entity_selector=HostSelector(), dimensions={counter.dimension_name: f"{counter.instance}"},
                            )
                    else:
                        value = aggregation_method([counter.value for counter in counters])
                        logger.info(f"Reporting value {value} for key {key}.")
                        logger.debug(f"The counters used for the calculation were: {counters}")
                        self.metric_methods[metric_type](key=key, value=value, entity_selector=HostSelector())

    def update_config(self):
        config_obj = {
            "openkit_beacon_url": self.config.get("openkit_beacon_url"),
            "openkit_application_id": self.config.get("openkit_application_id"),
            "openkit_proxy_address": self.config.get("openkit_proxy_address"),
            "openkit_proxy_username": self.config.get("openkit_proxy_username"),
            "openkit_proxy_password": self.config.get("openkit_proxy_password"),
            "openkit_use_oneagent_endpoints": self.config.get("openkit_use_oneagent_endpoints"),
            "openkit_verify_certificates": self.config.get("openkit_verify_certificates"),
            "openkit_capture_user_data": self.config.get("openkit_capture_user_data"),
            "log_level": self.config.get("log_level"),
        }
        config_file = os.path.join(current_file_path, "citrixagent", "config", "config.json")
        with open(config_file, "w") as f:
            logger.info(f"Writing config.json")
            json.dump(config_obj, f, indent=2)

    def process_log(self):
        try:
            self.times_not_logged_lines += 1
            for line in get_lines_from_file(self.log_path, 1000):
                self.times_not_logged_lines = 0
                self.logger.info(line)
                if "WARN" in line or "ERROR" in line:
                    self.results_builder.report_custom_info_event(
                        f"Got a bad log line from the citrix agent: {line}", "Bad log line from citrixagent.exe", entity_selector=HostSelector()
                    )
                if "FATAL" in line:
                    self.results_builder.report_error_event(
                        f"Fatal error running the citrix agent: {line}", "Fatal error from citrixagent.exe", entity_selector=HostSelector()
                    )

        except Exception as e:
            logger.error(f"Could not get citrixagent.exe logs: {e}")

    def report_vda_properties(self):
        vda_file_path = os.path.join(current_file_path, "citrixagent", "config", "vda.json")
        try:
            with open(vda_file_path, "r") as f:
                json_contents = json.load(f)
                for content in json_contents:
                    string_props = content.get("StringProperties", {})
                    self._report_property("Citrix broker version", string_props.get("BrokerVersion"))
                    self._report_property("Citrix desktop catalog", string_props.get("DesktopCatalogName"))
                    self._report_property("Citrix farm name", string_props.get("FarmName"))
                    self._report_property("Citrix desktop group", string_props.get("DesktopGroupName"))
                    self._report_property("Citrix desktop ID", string_props.get("Desktopid"))
        except Exception as ex:
            logger.warning(f"Could not process vda.json: {ex}")

    def _report_property(self, key: str, value: str):
        self.results_builder.add_property(PluginProperty(key, value, HostSelector(), MEAttribute.CUSTOM_HOST_METADATA))
